# coding=utf-8
# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .fp16util import (
    BN_convert_float,
    network_to_half,
    prep_param_lists,
    model_grads_to_master_grads,
    master_params_to_model_params,
    tofp16,
    to_python_float,
    clip_grad_norm,
    convert_module,
    convert_network,
    FP16Model,
)

from .fp16 import *
from .loss_scaler import *
